// VERSION: 0.3
// CREDITS: SamyGo <http://samygo.sourceforge.net/>
// CREDITS: geo650 (universal injection), moras86 (new icon), reaper7 (SDL/ttf code)
// WARNING: Tested only on LExxB650T2W with T-CHL7DEUC 2005.0 firmware;
// WARNING: devel/alpha/test level code; use at your own risk !!!
// LICENCE: GPL/CC

#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <SDL/SDL.h>
#include <SDL/SDL_ttf.h>
#include <libconfig.h>

#include "sd_video.h"

#define KEY_NOTHING 0x1c
#define FHD_W 1920
#define FHD_H 1080

extern int _ZN9KeyCommon17SendKeyPressInputEii(int, int, int);

static char gm_path[256];
static int mod_init=0;
static int key_activate=0x16; // KEY_16_COL_BLUE
static int res_sanity_check=2;
static int active_screen=0;

static int rr_on=0;
static int rr_acnt=0;

static int g_step_x=16;
static int g_step_y=9;

static int c_step_x=8;
static int c_step_y=5;

struct g_ar_node {
	int x; int y; int w; int h;
	char *label;
};

static struct g_ar_node ar_tble[] = {
   { 284, 0, 1350, 1080, "5:4" },
   { 240, 0, 1440, 1080, "4:3" },
   { 150, 0, 1620, 1080, "3:2" },
   { 96, 0, 1728, 1080, "16:10" },
   { 0, 0, 1920, 1080, "16:9 (native)" },
   { 0, 20, 1920, 1038, "1.85:1" },
   { 0, 132, 1920, 816, "2.35:1" },
   { 0, 138, 1920, 804, "2.39:1" },
   { 0, 192, 1920, 696, "2.76:1" },
};

static int ar_tb_index=4;
static int ar_tb_max_idx=8;

static struct g_ar_node cg_tble[] = {
   { 0, 0, 1920, 1080, "Custom (1)" },
   { 0, 0, 1920, 1080, "Custom (2)" },
   { 0, 0, 1920, 1080, "Custom (3)" },
   { 0, 0, 1920, 1080, "Custom (4)" },
   { 0, 0, 1920, 1080, "Custom (5)" },
   { 0, 0, 1920, 1080, "Custom (6)" },
   { 0, 0, 1920, 1080, "Custom (7)" },
   { 0, 0, 1920, 1080, "Custom (8)" },
   { 0, 0, 1920, 1080, "Custom (9)" },
   { 0, 0, 1920, 1080, "Custom (10)" },
   { 0, 0, 1920, 1080, "Custom (11)" },
   { 0, 0, 1920, 1080, "Custom (12)" },
   { 0, 0, 1920, 1080, "Custom (13)" },
   { 0, 0, 1920, 1080, "Custom (14)" },
   { 0, 0, 1920, 1080, "Custom (15)" },
   { 0, 0, 1920, 1080, "Custom (16)" },
};

static int cg_tb_index=0;
static int cg_tb_max_idx=-1;

static int _sdvideo_get (int sd_i, SdVideo_Settings_t *p) {
   memset(p, 0, sizeof(SdVideo_Settings_t));
   int res=_Z11SdVideo_Get11SdVideoID_kP18SdVideo_Settings_t(sd_i, p);
   if (!res) return(0);
   printf("SD(): ERROR: SdVideo_Get(%i) return=%d\n", sd_i, res);
   return(-1);
}

static int _sdvideo_chk (SdVideo_Settings_t *p, int hR, int vR) {
   if (p->bVideoOn != 1) return(-1);
   if (!active_screen && p->bGameModeOn != 0 && p->bGameModeOn != 1) return(-1);
   if (p->capt_hStart < 0 || p->capt_vStart < 0 || p->capt_hResolution < hR || p->capt_vResolution < vR) return(-1);
   return(0);
}

static void _sd_settings_dump (int sd_i) {
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   
   if (_sdvideo_get(sd_i, p)) return;

   printf("\n############################################################\n###\n");
   printf("### SdVideo_Get(%d: %s) SETTINGS\n", sd_i, (sd_i? "PIP": "MAIN"));
   printf("###\n############################################################\n\n");
   // "sScaleData        : %d,%d,%d,%d, %dx%d@%d%c, %dx%d"
   printf("\n[ CAPTURE ]\n\n");
   printf("\thStart: %d\n\tvStart: %d\n\tvResolution: %d\n\thResolution: %d\n", p->capt_hStart, p->capt_vStart, p->capt_hResolution, p->capt_vResolution);
   printf("\thActive: %d\n\tvActive: %d\n\thTotal: %d\n\tvTotal: %d\n", p->capt_hActive, p->capt_vActive, p->capt_hTotal, p->capt_vTotal);
   printf("\tbProgressScan: %d\n\tframeRate: %d\n\tbHdmiPC: %d\n\tbFrameDoubling: %d\n", p->bProgressScan, p->frameRate, p->bHdmiPC, p->bFrameDoubling);

   printf("\n[ GEOMETRY ]\n\n");
   printf("\tX: %d\n\tY: %d\n\tWidth: %d\n\tHeight: %d\n", p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height);

   printf("\n[ OTHERS ]\n\n");
   printf("\tBGColor: R=%d G=%d B=%d\n", p->sBGColor_R, p->sBGColor_G, p->sBGColor_B);
   printf("\tScaler: %s\n", ((p->bVideoOn)? "On": "Off"));
   printf("\t3D pull down: %d\n\tbGameModeOn: %d\n\tb3DGameModeOn: %d\n\tVideoDelay: %d\n", p->bEnable32PullDown, p->bGameModeOn, p->b3DGameModeOn, p->VideoDelay);

   // 0..5 { RGB, YUV420, YUV422, YUV444, XVYCC, UNKNOWN }
   printf("\teColorFormat: %d\n", p->eColorFormat);
   // 0..8 { SD_COR_SYS_AUTO, PAL, PAL60, PAL_M, PAL_N, SECAM, NT358, NT443, SD_COLOR_SYS_MAX }
   printf("\teCorSys: %d\n", p->eCorSys);
   printf("\tunkn_68: %d\n", p->unkn_68);

   printf("\n[ BG / MHEG ]\n\n");
   printf("\tbMHEGScale: %d\n\tbGfxScale: %d\n", p->MHEG_bMHEGScale, p->MHEG_bGfxScale);
   printf("\tsCrop(x,y,w,h): (%d,%d,%d,%d)\n", p->MHEG_sCrop_x, p->MHEG_sCrop_y, p->MHEG_sCrop_w, p->MHEG_sCrop_h);
   printf("\tsGeometry(x,y,w,h): (%d,%d,%d,%d)\n", p->MHEG_sGeometry_x, p->MHEG_sGeometry_y, p->MHEG_sGeometry_w, p->MHEG_sGeometry_h);
   printf("\tsScreen(x,y,w,h): (%d,%d,%d,%d)\n", p->MHEG_sScreen_x, p->MHEG_sScreen_y, p->MHEG_sScreen_w, p->MHEG_sScreen_h);
   printf("\tunkn_a4[0,1]: [%d,%d]\n", p->unkn_a4[0], p->unkn_a4[1]);
   printf("\n");
}

static int _sd_res_check(SdVideo_Settings_t *p) {
   int x=p->sGe_x; int y=p->sGe_y; int w=p->sGe_width; int h=p->sGe_height;

   if (x < 0 || (x > (FHD_W-320)) || y < 0 || (y > (FHD_H-200))) return(-1);
   if (w < 0 || w > FHD_W || h < 0 || h > FHD_H) return(-1);
   if ( ((x+w) > FHD_W) || ((y+h) > FHD_H) ) return(-1);
   if (w < 320 || h < 200) return(-1);
   
   return(0);
}

static int _sdl_txt(char *msg, int msec, int fsize) {
   SDL_Surface *screen;
   SDL_Surface *btsu;
   char *fpath="/mtd_appdata/Font/shadow.ttf";
   struct stat sts;
   TTF_Font *font;
   SDL_Color c_black={ 0, 0, 0 };
   SDL_Color c_yellow={ 255, 255, 0 };
   SDL_Rect dstrect = { 20, 20, 0, 0 };

   if (stat(fpath, &sts)) return(-1);
   if (SDL_Init(SDL_INIT_VIDEO)) return(-1);
   screen=SDL_SetVideoMode(FHD_W/2, FHD_H/2, 32, SDL_HWSURFACE);
   if (screen == NULL) return(-1);

   if (fsize == 0) { fsize=24; };
   if (TTF_Init() || ((font=TTF_OpenFont(fpath, fsize)) == NULL)) {
      SDL_Quit(); return(-1);
   }
   
   btsu=TTF_RenderUTF8_Shaded(font, msg, c_yellow, c_black);
   SDL_BlitSurface(btsu, NULL, screen, &dstrect);

   SDL_FreeSurface(btsu);
   SDL_Flip(screen);

   usleep(1000*msec);

   TTF_CloseFont(font);
   SDL_Quit();
   return(0);
}

static void _no_input_osd() {
   char mtxt[80];

   sprintf(mtxt, " No valid input on screen \"%s\" ... ", (active_screen? "PIP": "MAIN"));
   _sdl_txt(mtxt, 1000, 0);
}

static int _sd_geometry_change(int sd_i, int stepx, int stepy) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h;
   // int min_w=1280; int min_h=720;
   int min_w=960; int min_h=540;
   char mtxt[200];
   float ar;

   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100) || _sd_res_check(p)) {
      _no_input_osd(); return(-1); }

   x=p->sGe_x; y=p->sGe_y; w=p->sGe_width; h=p->sGe_height;
   if (sd_i) { min_w=320; min_h=200; };
   x-=stepx; if (x < 0) x=0; if (x > (FHD_W-min_w)) x=FHD_W-min_w;
   y-=stepy; if (y < 0) y=0; if (y > (FHD_H-min_h)) y=FHD_H-min_h;
   w+=2*stepx; if (w < min_w) { w=min_w; x=p->sGe_x; }; if ((x+w) > FHD_W) w=FHD_W-x;
   h+=2*stepy; if (h < min_h) { h=min_h; y=p->sGe_y; }; if ((y+h) > FHD_H) h=FHD_H-y;
   ar=((float)w/(float)h);
   if (res_sanity_check && (ar < 1.0 || ar > 2.8)) {
      sprintf(mtxt, "Aspect ratio '%.02f' out of bounds !!! ", ar);
      _sdl_txt(mtxt, 700, 0);
      return(-1);
   }

   printf("SD(G): [%d,%d,%d,%d] -> [%d,%d,%d,%d]\n", p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height, x, y, w, h);
   if (p->sGe_x == x && p->sGe_y == y && p->sGe_width == w && p->sGe_height == h) {
      sprintf(mtxt, " AR %.02f:1  [%d,%d,%d,%d] (not changed) ", ar, p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height);
      _sdl_txt(mtxt, 700, 0);
      return(0);
   }
   
   sprintf(mtxt, " [G] AR %.02f:1  [%d,%d,%d,%d] -> [%d,%d,%d,%d] ", ar, p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height, x, y, w, h);
   p->sGe_x=x; p->sGe_y=y; p->sGe_width=w; p->sGe_height=h;
   if (_sd_res_check(p)) {
      printf("SD(G): ERROR: parameter[s] out of range !!!\n"); return(-1); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(G): ERROR: SdVideo_Set() return=%d\n !!!", res); return(-1); }

   _sdl_txt(mtxt, 700, 0);
   return(0);
}

static int _sd_geometry_move(int sd_i, int dx, int dy) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h;
   char mtxt[200];

   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100) || _sd_res_check(p)) return(-1);

   x=p->sGe_x; y=p->sGe_y; w=p->sGe_width; h=p->sGe_height;
   if (dx > 0 && ((x+dx+w) > FHD_W)) dx=FHD_W-(x+w);
   if (dy > 0 && ((y+dy+h) > FHD_H)) dy=FHD_H-(y+h);
   if (dx < 0 && ((x+dx) < 0)) dx=0;
   if (dy < 0 && ((y+dy) < 0)) dy=0;
   x+=dx; y+=dy;
   
   if (p->sGe_x == x && p->sGe_y == y) return(0);
   
   printf("SD(M): [%d,%d] -> [%d,%d]\n", p->sGe_x, p->sGe_y, x, y);
   sprintf(mtxt, " MOVE [%d,%d] -> [%d,%d] ", p->sGe_x, p->sGe_y, x, y);
   p->sGe_x=x; p->sGe_y=y;
   if (_sd_res_check(p)) {
      printf("SD(M): ERROR: parameter[s] out of range !!!\n"); return(-1); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(M): ERROR: SdVideo_Set() return=%d\n !!!", res); return(-1); }

   _sdl_txt(mtxt, 500, 0);
   return(0);
}

static int _sd_geometry_set(int sd_i, int tx) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h;
   char mtxt[200];
   float ar;
   struct g_ar_node gstr;

   if (sd_i) {
      _sdl_txt(" No predefined settings for screen \"PIP\" ", 1000, 0);
      return(-1); }
   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100) || _sd_res_check(p)) {
      _no_input_osd(); return(-1); }
   if (tx > ar_tb_max_idx) return(-1);
   
   gstr=ar_tble[tx];
   x=gstr.x; y=gstr.y; w=gstr.w; h=gstr.h; ar=((float)w/(float)h);

   printf("SD(G): [%d,%d,%d,%d] -> [%d,%d,%d,%d]\n", p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height, x, y, w, h);
   sprintf(mtxt, " Mode %s ", gstr.label);

   if (p->sGe_x == x && p->sGe_y == y && p->sGe_width == w && p->sGe_height == h) {
      sprintf(mtxt, " AR %.02f:1  [%d,%d,%d,%d] (not changed) ", ar, p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height);
      _sdl_txt(mtxt, 700, 0);
      return(0);
   }
   
   p->sGe_x=x; p->sGe_y=y; p->sGe_width=w; p->sGe_height=h;
   if (_sd_res_check(p)) {
      printf("SD(G): ERROR: parameter[s] out of range !!!\n"); return(-1); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(G): ERROR: SdVideo_Set() return=%d\n !!!", res); return(-1); }

   ar_tb_index=tx;
   _sdl_txt(mtxt, 1000, 36);
   return(0);
}

static int _sd_geometry_original(int sd_i, int center_flg) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h,w_act,h_act;
   char mtxt[200];
   float ar;
   
   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100)) {
      _no_input_osd(); return(-1); }

   x=p->sGe_x; y=p->sGe_y; w=p->sGe_width; h=p->sGe_height; w_act=p->capt_hActive; h_act=p->capt_vActive;
   if (w_act < 160 || w_act > FHD_W || h_act < 100 || h_act > FHD_H) return(-1);
   ar=((float)w_act/(float)h_act);
   w=w_act; h=h_act;
   if (center_flg) { x=(FHD_W-w)/2; y=(FHD_H-h)/2; } else { x=0; y=0; }
   printf("SD(G): [%d,%d,%d,%d] -> [%d,%d,%d,%d]\n", p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height, x, y, w, h);
   sprintf(mtxt, " Mode: %dx%d (original, %s) AR=%.02f ", w_act, h_act, (center_flg? "centered": "left/top"), ar);
   
   if (p->sGe_x == x && p->sGe_y == y && p->sGe_width == w && p->sGe_height == h) {
      sprintf(mtxt, " AR %.02f:1  [%d,%d,%d,%d] (not changed) ", ar, p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height);
      _sdl_txt(mtxt, 700, 0);
      return(0);
   }

   p->sGe_x=x; p->sGe_y=y; p->sGe_width=w; p->sGe_height=h;
   if (_sd_res_check(p)) {
      printf("SD(G): ERROR: parameter[s] out of range !!!\n"); return(-1); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(G): ERROR: SdVideo_Set() return=%d\n !!!", res); return(-1); }

   _sdl_txt(mtxt, 1500, 36);
   return(0);
}

static int _sd_capture_change(int sd_i, int stepx, int stepy) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h,w_act,h_act;
   int min_w=240; int min_h=160;
   char mtxt[200];
   float ar;
   
   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, min_w, min_h) || _sd_res_check(p)) {
      _no_input_osd(); return(-1); }
   if (p->capt_hActive < min_h || p->capt_vActive < min_w) return(-1);
   
   x=p->capt_hStart; y=p->capt_vStart; w=p->capt_hResolution; h=p->capt_vResolution;
   w_act=p->capt_hActive; h_act=p->capt_vActive;
   
   x-=stepx; if (x < 0) x=0; if (x > (w_act-min_w)) x=w_act-min_w;
   y-=stepy; if (y < 0) y=0; if (y > (h_act-min_h)) y=h_act-min_h;
   w+=2*stepx; if (w < min_w) { w=min_w; x=p->capt_hStart; }; if ((x+w) > w_act) w=w_act-x;
   h+=2*stepy; if (h < min_h) { h=min_h; y=p->capt_vStart; }; if ((y+h) > h_act) h=h_act-y;
   ar=((float)w/(float)h);
      
   printf("SD(C): [%d,%d,%d,%d] -> [%d,%d,%d,%d]\n", p->capt_hStart, p->capt_vStart, p->capt_hResolution, p->capt_vResolution, x, y, w, h);

   if (p->capt_hStart == x && p->capt_vStart == y && p->capt_hResolution == w && p->capt_vResolution == h) {
      sprintf(mtxt, " [C] AR=%.02f CROP=[%d,%d,%d,%d] (not changed) ", ar, x, y, w, h);
      _sdl_txt(mtxt, 700, 0);
      return(0);
   }
   
   sprintf(mtxt, " [C] ACV=%dx%d AR=%.02f [%d,%d,%d,%d] -> [%d,%d,%d,%d] ", w_act, h_act, ar,
      p->capt_hStart, p->capt_vStart, p->capt_hResolution, p->capt_vResolution, x, y, w, h);

   p->capt_hStart=x; p->capt_vStart=y; p->capt_hResolution=w; p->capt_vResolution=h;
   if (_sd_res_check(p)) {
      printf("SD(C): ERROR: parameter[s] out of range !!!\n"); return(-1); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(C): ERROR: SdVideo_Set() return=%d\n !!!", res); return(-1); }

   _sdl_txt(mtxt, 700, 0);
 
   return(0);
}

static int _sd_geometry_custom(int sd_i, int tx) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h;
   char mtxt[200];
   float ar;
   struct g_ar_node gstr;

   if (sd_i) {
      _sdl_txt(" No custom settings for screen \"PIP\" ", 1000, 0);
      return(-1); }
   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100) || _sd_res_check(p)) {
      _no_input_osd(); return(-1); }
   if (tx > cg_tb_max_idx) return(-1);

   gstr=cg_tble[tx];
   x=gstr.x; y=gstr.y; w=gstr.w; h=gstr.h; ar=((float)w/(float)h);

   printf("SD(CG): [%d,%d,%d,%d] -> [%d,%d,%d,%d]; IDX=%d\n", p->sGe_x, p->sGe_y, p->sGe_width, p->sGe_height, x, y, w, h, tx);
   sprintf(mtxt, " User defined [%d,%d,%d,%d] :  %s", x, y, w, h, gstr.label);

   if (p->sGe_x == x && p->sGe_y == y && p->sGe_width == w && p->sGe_height == h) {
      _sdl_txt(mtxt, 700, 0);
      return(0); }
   
   p->sGe_x=x; p->sGe_y=y; p->sGe_width=w; p->sGe_height=h;
   if (_sd_res_check(p)) {
      sprintf(mtxt, " Setting \"%d\": parameter out of range !!! ", tx+1);
      _sdl_txt(mtxt, 1000, 0);
      printf("SD(CG): ERROR: parameter[s] out of range !!!\n"); return(0); }

   if (res=_Z11SdVideo_Set11SdVideoID_kP18SdVideo_Settings_t(sd_i, p)) {
      printf("SD(CG): ERROR: SdVideo_Set() return=%d\n !!!", res); return(0); }

   _sdl_txt(mtxt, 1000, 0);
   return(0);
}

static int _sd_osd_info (int sd_i) {
   int res;
   SdVideo_Settings_t s;
   SdVideo_Settings_t *p=&s;
   int x,y,w,h;
   char mtxt[200];
   float ar;

   if (_sdvideo_get(sd_i, p) || _sdvideo_chk(p, 160, 100)) {
      _no_input_osd(); return(-1); }

   x=p->sGe_x; y=p->sGe_y; w=p->sGe_width; h=p->sGe_height; ar=((float)w/(float)h);
   sprintf(mtxt, " [%s] AR %.02f:1 IN=%dx%d%s CROP=[%d,%d,%d,%d] OUT=[%d,%d,%d,%d]",
      (active_screen? "PIP": "MAIN"),
      ar, p->capt_hActive, p->capt_vActive,
      ((p->bProgressScan)? "p": "i"), p->capt_hStart, p->capt_vStart, p->capt_hResolution, p->capt_vResolution,
      x, y, w, h);
   
   _sdl_txt(mtxt, 2000, 22);
   return(0);
}

static void _screen_switch() {
   char mtxt[80];

   active_screen=1-active_screen;
   // todo: change step[xy] etc
   sprintf(mtxt, " Screen switch: %s ", (active_screen? "PIP": "MAIN"));
   _sdl_txt(mtxt, 1000, 36);
}

static int _init_mod_func() {
   char sbuf[256];
   struct stat sts;
   config_t rcfg;
   config_setting_t *setting;
   const char *vstr;
   int ival;

   sprintf(sbuf, "%sresizer.cfg", gm_path);
   if (stat(sbuf, &sts)) {
      printf("SD(config): ERROR: file '%s' not found\n", sbuf);
      _sdl_txt(" ERROR: Cant find config file !!! ", 2000, 0);
      return(-1); }

   config_init(&rcfg);
   if (!config_read_file(&rcfg, sbuf)) {
      printf("SD(config): ERROR: %s:%d - %s\n", config_error_file(&rcfg), config_error_line(&rcfg), config_error_text(&rcfg));
      _sdl_txt(" ERROR: Config file syntax error (?) ", 2000, 0);
      config_destroy(&rcfg);
      return(-1); }

   if (config_lookup_string(&rcfg, "version", &vstr)) {
      printf("SD(config): found magic version string '%s'\n", vstr);
   } else {
      printf("SD(config): ERROR: magic vstring not found in config file\n");
      config_destroy(&rcfg);
      return(-1);
   }

   if (config_lookup_int(&rcfg, "key_activate", &ival) && ival > 0 && ival < 256) {
      printf("SD(config): key_activate = %d  (0x%02X)\n", ival, ival);
      key_activate=ival;
   }

   while (1) {
      setting=config_lookup(&rcfg, "custom_geometry");
      if (setting == NULL || !config_setting_is_list(setting)) break;
      int cgc=config_setting_length(setting);
      if (cgc < 1) break; if (cgc > 16) cgc=16;
      printf("SD(config): %d custom geometry settings\n", cgc);

      int i; for (i=0; i < cgc; i++) {
         config_setting_t *item=config_setting_get_elem(setting, i);
         if (item == NULL || !config_setting_is_array(item) || config_setting_length(item) != 4) continue;
         
         int x,y,w,h;
         x=config_setting_get_int_elem(item, 0);
         y=config_setting_get_int_elem(item, 1);
         w=config_setting_get_int_elem(item, 2);
         h=config_setting_get_int_elem(item, 3);

         if (x < 0 || (x > (FHD_W-320)) || y < 0 || (y > (FHD_H-200))) continue;
         if (w < 0 || w > FHD_W || h < 0 || h > FHD_H) continue;
         if (((x+w) > FHD_W) || ((y+h) > FHD_H)) continue;
         if (w < 320 || h < 200) continue;

         cg_tb_max_idx++;
         cg_tble[cg_tb_max_idx].x=x;
         cg_tble[cg_tb_max_idx].y=y;
         cg_tble[cg_tb_max_idx].w=w;
         cg_tble[cg_tb_max_idx].h=h;
         printf("SD(config): setting=%d cg[%d]=[%d,%d,%d,%d]\n", i, cg_tb_max_idx, x, y, w, h);
      }
      break;
   }
   
   config_destroy(&rcfg);
   return(1);
}

static int _adwtfunc_call_local(int key) __attribute__ ((noinline));
static int _adwtfunc_call_local(int key) {
   int sd_i=active_screen;

   if (mod_init == -1) return(key);
   if (!rr_on) {
      if (key == key_activate) {
         rr_acnt++; } else { rr_acnt=0; }
      if (rr_acnt < 2) return(key);
      if (mod_init == 0) mod_init=_init_mod_func();
      if (mod_init == -1) return(KEY_NOTHING);
      rr_on=1; rr_acnt=0;
      _sdl_txt(" Resize mode: ON ", 500, 36);
      return(KEY_NOTHING);
   }
   if (key == key_activate) {
      rr_acnt=0; rr_on=0;
      _sdl_txt(" Resize mode: OFF ", 500, 36);
      return(KEY_NOTHING);
   }
   
   if (key == 0x1f) { // KEY_1F_INFO
      _sd_osd_info(sd_i);
      return(KEY_NOTHING);
   }
   if (key == 0x09) { // "5"
      int i; for (i=0; i <= 1; i++) _sd_settings_dump(i);
      return(KEY_NOTHING);
   }

   // geometry: move (pan)
   if (key == 0x08) { // "4"
      _sd_geometry_move(sd_i, -g_step_x, 0);
      return(KEY_NOTHING);
   }   
   if (key == 0x0a) { // "6"
      _sd_geometry_move(sd_i, g_step_x, 0);
      return(KEY_NOTHING);
   }
   if (key == 0x05) { // "2"
      _sd_geometry_move(sd_i, 0, -g_step_y);
      return(KEY_NOTHING);
   }   
   if (key == 0x0d) { // "8"
      _sd_geometry_move(sd_i, 0, g_step_y);
      return(KEY_NOTHING);
   }
      
   // original (1:1 pixel mapping)
   if (key == 0x47) { // KEY_47_PLAY
      _sd_geometry_original(sd_i, 1);
      return(KEY_NOTHING);
   }
   if (key == 0x46) { // KEY_46_STOP
      _sd_geometry_original(sd_i, 0);
      return(KEY_NOTHING);
   }

   // set geometry (predefined sets of values)
   if (key == 0x45) { // KEY_45_FB [<<]
      _sd_geometry_set(sd_i, ((ar_tb_index <= 0)? ar_tb_max_idx: (ar_tb_index-1)));
      return(KEY_NOTHING);
   }
   if (key == 0x48) { // KEY_48_FF [>>]
      _sd_geometry_set(sd_i, ((ar_tb_index >= ar_tb_max_idx)? 0: (ar_tb_index+1)));
      return(KEY_NOTHING);
   }

   // change geometry
   if (key == 0x07) { // KEY_07_VOLUME_UP
      _sd_geometry_change(sd_i, g_step_x, 0);
      return(KEY_NOTHING);
   }
   if (key == 0x0b) { // KEY_0B_VOLUME_DOWN
      _sd_geometry_change(sd_i, -g_step_x, 0);
      return(KEY_NOTHING);
   }
   if (key == 0x12) { // KEY_12_P_UP
      _sd_geometry_change(sd_i, 0, g_step_y);
      return(KEY_NOTHING);
   }
   if (key == 0x10) { // KEY_10_P_DOWN
      _sd_geometry_change(sd_i, 0, -g_step_y);
      return(KEY_NOTHING);
   }

   // change capture
   if (key == 0x62) { // KEY_62_RIGHT
      _sd_capture_change(sd_i, -c_step_x, 0);
      return(KEY_NOTHING);
   }
   if (key == 0x65) { // KEY_65_LEFT
      _sd_capture_change(sd_i, c_step_x, 0);
      return(KEY_NOTHING);
   }
   if (key == 0x60) { // KEY_60_UP
      _sd_capture_change(sd_i, 0, -c_step_y);
      return(KEY_NOTHING);
   }
   if (key == 0x61) { // KEY_61_DOWN
      _sd_capture_change(sd_i, 0, c_step_y);
      return(KEY_NOTHING);
   }
   
   // picture switch (main, pip)
   if (key == 0x11) { // "0"
      _screen_switch();
      return(KEY_NOTHING);
   }

   // custom geometry settings toggle
   if (key == 0x1a) { // "KEY_1A_MENU"
      if (cg_tb_max_idx < 0 || _sd_geometry_custom(sd_i, cg_tb_index)) return(KEY_NOTHING);
      cg_tb_index++; if (cg_tb_index > cg_tb_max_idx) cg_tb_index=0;
      return(KEY_NOTHING);
   }

   return(key);
}


// universal injection procedure by geo650
// copied from mybuttons.c by reaper7
void mod_sd_video_injection(int a, int key, int arg_r2) __attribute__ ((noinline));
void mod_sd_video_injection(int a, int key, int arg_r2)
{
	// CAUTION! Do not add/modify code here... or modify injection indexes in loader

	// [0][1] gcc should put "STMFD SP!, {R4,LR}" and "MOV R4,[R1]" here
	asm volatile(
		"STMFD   SP!, {R0,R2,R3,R12}\n"		// [2] store some registers
	);

	// first, we have to execute our function
	_adwtfunc_call_local(key);				// [3][4] calling local function with key variable in R0, returning R0

	// then, prepare for calling and call previous injection function
	asm volatile(
		"MOV     R1, R0\n"			// [5] remapped keycode
		"MOV     R6, R0\n"			// [6] remapped keycode
		"LDMFD   SP!, {R0,R2,R3,R12}\n"		// [7] restore registers stored before
		"LDMFD   SP!, {R4,LR}\n"		// [8] restore registers...
		"MOV     R7, R2\n"			// [9] This is original command that we overwrite (and replace at next injection)
		"LDR     R3, [R8]\n"			// [10] This is original command that we overwrite (and replace at next injection)
		"LDR     PC, =(_ZN9KeyCommon17SendKeyPressInputEii + 0x18 )\n"	// [11] return after injection (not reached at next injection)
	);
	// and that's all; the rest (returning to exeDSP) is already programmed in the first injected function
}

// setting Game_Main path
void mod_sd_video_set_path(char *p)
{
   strcpy(gm_path, p);
}
